--[[ Crafty
     Crafting daily helper.

]]--

local addoninfo, crafty = ...

-- item detail cache for name and rarity
crafty.details = {}

local html_safe = Library.printf.html_safe
local printf = Library.printf.printf
local printfh = Library.printf.printfh
local printfs = Library.printf.printfs
local printfhs = Library.printf.printfhs
local sprintf = Library.printf.sprintf
local sprintfh = Library.printf.sprintfh
local cron = Library.LibCron

crafty.debugging = false

crafty.watching_quests = false

local color_scale = 0.1
crafty.clearer = { colorR = color_scale, colorB = color_scale, colorG = color_scale, strength = 2.5, blurX = 6, blurY = 6 }

function crafty.variables(event, addon)
  if addon == 'Crafty' then
    Crafty_sv = Crafty_sv or { }
    Crafty_sv.loc = Crafty_sv.loc or { x = 0, y = 500 }
    Crafty_sv.fontsize = Crafty_sv.fontsize or 14
    Crafty_sv.verbose = Crafty_sv.verbose or false
    Crafty_sv.show_replaced = Crafty_sv.show_replaced or true
    if Crafty_sv.reagents_first == nil then
      Crafty_sv.reagents_first = true
    end
    if Crafty_sv.autoshow == nil then
      Crafty_sv.autoshow = true
    end
    Crafty_sv.max_items = Crafty_sv.max_items or 10
  end
  if crafty.verbose == nil then
    crafty.verbose = Crafty_sv.verbose
  else
    -- may have been set by the user somehow before the variables were loaded?
    Crafty_sv.verbose = crafty.verbose
  end
  if crafty.show_replaced == nil then
    crafty.show_replaced = Crafty_sv.show_replaced
  else
    -- may have been set by the user somehow before the variables were loaded?
    Crafty_sv.show_replaced = crafty.show_replaced
  end
  if crafty.autoshow == nil then
    crafty.autoshow = Crafty_sv.autoshow
  else
    -- may have been set by the user somehow before the variables were loaded?
    Crafty_sv.autoshow = crafty.autoshow
  end
  if crafty.autoshow then
    crafty.enable_quest_hooks()
  else
    crafty.disable_quest_hooks()
  end
end

function crafty.quest_hook(handle, quests)
  local we_care = false
  if not quests then
    return
  end
  for k, v in pairs(quests) do
    if crafty.quests[k] then
      we_care = true
      break
    end
  end
  if we_care then
    local details = crafty.quest_items_wanted()
    if next(details) then
      crafty.toggle_ui(true)
    else
      crafty.toggle_ui(false)
    end
  end
end

function crafty.enable_quest_hooks()
  if not crafty.watching_quests then
    crafty.watching_quests = true
    Command.Event.Attach(Event.Quest.Accept, crafty.quest_hook, "crafty quest hook")
    Command.Event.Attach(Event.Quest.Abandon, crafty.quest_hook, "crafty quest hook")
    Command.Event.Attach(Event.Quest.Change, crafty.quest_hook, "crafty quest hook")
    Command.Event.Attach(Event.Quest.Complete, crafty.quest_hook, "crafty quest hook")
    crafty.quest_hook({}, Inspect.Quest.List())
  end
end

function crafty.disable_quest_hooks()
  -- harmless to try to remove a hook that isn't attached
  Command.Event.Detach(Event.Quest.Accept, crafty.quest_hook)
  Command.Event.Detach(Event.Quest.Abandon, crafty.quest_hook)
  Command.Event.Detach(Event.Quest.Change, crafty.quest_hook)
  Command.Event.Detach(Event.Quest.Complete, crafty.quest_hook)
  crafty.watching_quests = false
end

-- remove any items we can from the list.
function crafty.mark_processed(objectives, processed_objectives, my_stuff)
  local did_any = false
  for k, v in pairs(objectives) do
    local item_type = crafty.items[k]
    if item_type and my_stuff[item_type] and my_stuff[item_type] > 0 then
      did_any = true
      local used = math.min(my_stuff[item_type], v.needed)
      if crafty.debugging then
        printf("Found %d/%d of %s (%d in inventory).", used, v.needed, crafty.name(k), my_stuff[item_type])
      end
      my_stuff[item_type] = my_stuff[item_type] - used
      objectives[k].needed = objectives[k].needed - used
      -- If we didn't find all of them, we'll be leaving the remainder in
      -- the wanted list, which could get them added to processed_objectives
      -- later.
      if not processed_objectives[k] then
	processed_objectives[k] = { needed = used, have = used, replaced = 0 }
      else
	processed_objectives[k].needed = processed_objectives[k].needed + used
	processed_objectives[k].have = processed_objectives[k].have + used
      end
      -- if there's nothing left to do here, drop it.
      if objectives[k].needed == 0 and objectives[k].replaced == 0 then
        objectives[k] = nil
      end
    end
  end
  return did_any
end

function crafty.get_details(key)
  if (not crafty.details[key]) and crafty.items[key] then
    local item_type = crafty.items[key]
    local status, value = pcall(Inspect.Item.Detail, item_type)
    if status then
      crafty.details[key] = value
    end
  end
end

function crafty.name(key)
  crafty.get_details(key)
  return (crafty.details[key] and crafty.details[key].name) or key
end

function crafty.rarity(key)
  crafty.get_details(key)
  return (crafty.details[key] and crafty.details[key].rarity) or 'unknown'
end

function crafty.is_reagent(key)
  crafty.get_details(key)

  return crafty.details[key] and crafty.details[key].category and crafty.details[key].category:find('reagent')
end

function crafty.find_recipes(objectives, new_objectives)
  local did_any = false
  for k, v in pairs(objectives) do
    if v.needed > 0 then
      if crafty.recipes[k] then
        local recipe = crafty.recipes[k]
        if crafty.debugging then
          printf("Found a recipe for %dx %s, replacing with components.", v.needed, recipe.name)
        end
        did_any = true
        for i = 1, #recipe.parts do
          local part = recipe.parts[i]
	  local needed = part.quantity * v.needed
	  if not new_objectives[part.key] then
	    new_objectives[part.key] = { needed = needed }
	  else
	    new_objectives[part.key].needed = new_objectives[part.key].needed + needed
	  end
        end
        -- mark it as needed, but also replaced, for later display
	objectives[k].replaced = (objectives[k].replaced or 0) + objectives[k].needed
        objectives[k].needed = 0
      end
    end
  end
  return did_any
end

function crafty.quest_items_wanted()
  -- what do we need?
  local crafting_quests = 0
  local quests = Inspect.Quest.List()
  local objectives = {}
  -- no quest info available?
  if not quests then
    return objectives
  end
  for k, v in pairs(quests) do
    local quest = crafty.quests[k]
    if quest then
      local details = Inspect.Quest.Detail(k)
      crafting_quests = crafting_quests + 1
      if crafty.debuggingg then
        printf("Found a known quest: %s", quest.name)
      end
      for i = 1, #quest.objectives do
	local o = quest.objectives[i]
	local needed = o.count
	-- we need to actually count the items that are "done" in some
	-- cases, such as if you have a gathering quest which wants lumber,
	-- and also need to make items which use that lumber.
	-- So we don't check objectivesDone, or the complete flag.
	if objectives[o.item] then
	  objectives[o.item].needed = objectives[o.item].needed + needed
	else
	  objectives[o.item] = { needed = needed, replaced = 0 }
	end
      end
    end
  end
  return objectives, crafting_quests
end

function crafty.compute()
  local processed_objectives = {}
  local objectives, crafting_quests = crafty.quest_items_wanted()
  local quests_to_do = 0
  -- special case: if there's nothing to do, don't do it
  if not next(objectives) then
    return { objectives = {}, crafting_quests = 0 }
  end
  -- okay, what have we got?
  local my_stuff = {}
  local inventories = { Utility.Item.Slot.Inventory(), Utility.Item.Slot.Vault(), Utility.Item.Slot.Quest(), Utility.Item.Slot.Bank() }
  for i = 1, #inventories do
    local stuff = Inspect.Item.Detail(inventories[i])
    for k, v in pairs(stuff) do
      local itype = v.type
      my_stuff[itype] = (my_stuff[itype] or 0) + (v.stack or 1)
    end
  end
  local count = 0
  local ok = true
  local any_replacements = false
  while ok do
    ok = false
    count = count + 1
    local new_objectives = {}
    -- step one: mark off things we have
    crafty.mark_processed(objectives, processed_objectives, my_stuff)
    -- step two: convert anything we didn't have to recipes.
    if crafty.find_recipes(objectives, new_objectives) then
      any_replacements = true
      -- it will be worth retrying
      ok = true
      -- step three: any items recipes wanted go back on the objective list.
      for k, v in pairs(new_objectives) do
	if crafty.debugging then
          printf("Adding %dx %s back into objectives.", v.needed, crafty.name(k))
	end
	if not objectives[k] then
	  objectives[k] = v
	  objectives[k].replaced = 0
	else
	  objectives[k].needed = objectives[k].needed + v.needed
	end
      end
    end
    -- then repeat, in case we have any of those ingredients handy,
    -- and if we don't, we can check again in case they can be crafted.
    if count > 5 then
      printf("Looped too much. Probably confused.")
      break
    end
  end
  -- now add anything we couldn't find, even with recipes, to
  -- processed_objectives
  local any_more = false
  for k, v in pairs(objectives) do
    if v.needed > 0 or v.replaced > 0 then
      any_more = true
    end
    if processed_objectives[k] then
      processed_objectives[k].needed = processed_objectives[k].needed + v.needed
      processed_objectives[k].replaced = (processed_objectives[k].replaced or 0) + (v.replaced or 0)
    else
      processed_objectives[k] = { needed = v.needed, have = 0, replaced = v.replaced }
    end
  end
  return {
    objectives = processed_objectives,
    any_more = any_more,
    any_replacements = any_replacements,
    crafting_quests = crafting_quests
  }
end

local function compare_objectives(a, b, processed_objectives)
  -- sort reagents first
  if crafty.is_reagent(a) and not crafty.is_reagent(b) then
    return Crafty_sv.reagents_first
  end
  if crafty.is_reagent(b) and not crafty.is_reagent(a) then
    return not Crafty_sv.reagents_first
  end
  -- alphabetical
  return crafty.name(a) < crafty.name(b)
end

function crafty.display_needed(data)
  local processed_objectives = data.objectives
  local any_needed = data.any_more
  local work_to_do = data.any_replacements
  local crafting_quests = data.crafting_quests
  local quests_to_work_on = 0
  local missing = false
  local lines = {}
  local keys = {}
  local displayed_dots = false
  for k, v in pairs(processed_objectives) do
    keys[#keys + 1] = k
  end
  table.sort(keys, function(a, b) return compare_objectives(a, b, processed_objectives) end)

  local non_quest_work = false
  for i = 1, #keys do
    local k = keys[i]
    local v = processed_objectives[k]
    local have, needed, replaced = v.have, v.needed, (v.replaced or 0)
    if needed > have or replaced > 0 then
      -- check for any non-quest items we need to show.
      if crafty.items[k] and crafty.rarity(k) ~= 'quest' then
        non_quest_work = true
	break
      end
    end
  end

  for i = 1, #keys do
    local k = keys[i]
    local v = processed_objectives[k]
    local name = "?(" .. k .. ")"
    local skip = false
    local have, needed, replaced = v.have, v.needed, (v.replaced or 0)
    if needed > 0 or (crafty.show_replaced and replaced > 0) then
      if crafty.items[k] then
	-- quest items:
	if crafty.rarity(k) == 'quest' then
	  if replaced > 0 then
	    quests_to_work_on = quests_to_work_on + 1
	  end
	  if non_quest_work then
	    -- don't show the sub-items until all the non-quest work, like
	    -- raw materials, is done.
	    skip = true
	  end
	end
	name = crafty.name(k)
      end
      if not crafty.verbose then
	local really_needed = crafty.show_replaced and (needed + replaced) or needed
        if have >= really_needed then
	  skip = true
	end
      end
      if not skip then
	local color = "#ddddaa"
	if have >= needed then
	  if replaced > 0 then
	    color = "#44bb44"
	  else
	    color = "#66dd66"
	  end
	else
	  -- reagents are flat grey
	  if crafty.is_reagent(k) then
	    color = "#aaaaaa"
	  end
	  missing = true
	end
	if (Crafty_sv.max_items > 0) and (#lines + 1 >= Crafty_sv.max_items) then
	  if not displayed_dots then
	    displayed_dots = true
	    lines[#lines + 1] = "[...]"
	  end
	else
	  lines[#lines + 1] = sprintfh("<font color=\"%s\">%d/%d %s</font>", color, have, needed + replaced, name)
	end
      end
    end
  end
  if not missing then
    if work_to_do then
      table.insert(lines, 1, sprintf("<font color=\"#88dd88\">Crafting to do: (Quests: %d/%d)</font>", quests_to_work_on or -1, crafting_quests or -1))
    else
      lines[#lines + 1] = sprintf("<font color=\"#66ee66\">Good to go! (Quests: %d)</font>", crafting_quests or -1)
    end
  else
    table.insert(lines, 1, sprintf("<font color=\"#b0b000\">Materials needed:</font> <font color=\"#aaaaaa\">(Quests: %d)</font>", crafting_quests or -1))
  end
  return lines
end

function crafty.create_ui()
  local x = Crafty_sv and Crafty_sv.loc.x or 0
  local y = Crafty_sv and Crafty_sv.loc.y or 500
  local fontsize = Crafty_sv and Crafty_sv.fontsize or 14
  crafty.context = UI.CreateContext("Crafty")
  crafty.context:SetStrata("hud")
  crafty.context:SetLayer(0)
  crafty.ui = UI.CreateFrame("Frame", "crafty_outer", crafty.context)
  crafty.ui:SetVisible(false)
  crafty.ui:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x, y)
  crafty.ui:SetBackgroundColor(0.7, 0.7, 0.7, 1.0)
  crafty.subframe = UI.CreateFrame("Frame", "crafty_inner", crafty.ui)
  crafty.subframe:SetPoint("TOPLEFT", crafty.ui, "TOPLEFT", 2, 2)
  crafty.subframe:SetBackgroundColor(0.1, 0.1, 0.1, 1.0)
  crafty.display_area = UI.CreateFrame("Text", "crafty_text", crafty.subframe)
  crafty.display_area:SetPoint("TOPLEFT", crafty.subframe, "TOPLEFT", 2, 2)
  crafty.display_area:SetEffectGlow(crafty.clearer)
  crafty.display_area:SetWidth(220)
  crafty.display_area:SetFontSize(fontsize)
  crafty.display_area:SetText("Crafty!")
  crafty.subframe:SetPoint("BOTTOMRIGHT", crafty.display_area, "BOTTOMRIGHT", 2, 2)
  crafty.ui:SetPoint("BOTTOMRIGHT", crafty.subframe, "BOTTOMRIGHT", 2, 2)
end

function crafty.update_hook()
  local data = crafty.compute()
  if data and data.objectives and next(data.objectives) then
    local lines = crafty.display_needed(data)
    if crafty.ui and crafty.ui:GetVisible() then
      crafty.display_area:SetText(table.concat(lines, "\n"), true)
      if needed then
        crafty.ui:SetBackgroundColor(0.8, 0.8, 0.6, 1.0)
      else
        crafty.ui:SetBackgroundColor(0.6, 0.9, 0.6, 1.0)
      end
    else
      printfhs("%s", html_safe(table.concat(lines, "\n")))
    end
  end
end

function crafty.enable_events()
  if not crafty.cron_hook then
    crafty.cron_hook = cron.new('Crafty', 1, true, false, crafty.update_hook)
  end
end

function crafty.disable_events()
  if crafty.cron_hook then
    cron.delete(crafty.cron_hook)
    crafty.cron_hook = false
  end
end

function crafty.toggle_ui(state)
  if not crafty.ui then
    crafty.create_ui()
  end
  if crafty.ui then
    if state == nil then
      state = not crafty.ui:GetVisible()
    end
    crafty.ui:SetVisible(state)
    if state then
      crafty.update_hook()
      crafty.enable_events()
    else
      crafty.disable_events()
    end
  end
  return state
end

function crafty.slashcommand(args)
  local did_something = false
  
  if not args then
    return
  end

  if not Crafty_sv then
    if args.a or args.x or args.y or args.f or args.s then
      printf("Can't run the slash command for GUI stuff before variables are loaded.")
      return
    end
  end

  if args.m then
    did_something = true
    if args.m >= 0 then
      Crafty_sv.max_items = args.m
      printfh("Max items: <font color=\"#ffffcc\">%d</font>", Crafty_sv.max_items)
    else
      printf("Maximum must be 0 (disabled) or higher.")
    end
  end

  if args.x then
    Crafty_sv.loc.x = args.x
    did_something = true
  end

  if args.y then
    Crafty_sv.loc.y = args.y
    did_something = true
  end

  if args.x or args.y then
    if crafty.ui then
      local x = Crafty_sv.loc.x or 0
      local y = Crafty_sv.loc.y or 500
      printf("x, y: %d, %d", x, y)
      crafty.ui:SetPoint("TOPLEFT", UIParent, "TOPLEFT", x, y)
    end
  end

  if args.f then
    Crafty_sv.fontsize = args.f
    if crafty.ui then
      crafty.display_area:SetFontSize(Crafty_sv.fontsize)
    end
    did_something = true
  end

  if args.d then
    crafty.debugging = not crafty.debugging
    printfh("Debugging: <font color=\"#ffffcc\">%s</font>", tostring(crafty.debugging))
    did_something = true
  end

  if args.r then
    Crafty_sv.reagents_first = not Crafty_sv.reagents_first
    printfh("Reagents first: <font color=\"#ffffcc\">%s</font>", tostring(Crafty_sv.reagents_first))
    did_something = true
  end

  if args.a then
    crafty.autoshow = not crafty.autoshow
    Crafty_sv.autoshow = crafty.autoshow
    printfh("Autoshow: <font color=\"#ffffcc\">%s</font>", tostring(crafty.autoshow))
    did_something = true
  end

  if args.v then
    crafty.verbose = not crafty.verbose
    Crafty_sv.verbose = crafty.verbose
    printfh("Verbosity: <font color=\"#ffffcc\">%s</font>", tostring(crafty.verbose))
    did_something = true
  end

  if args.V then
    crafty.show_replaced = not crafty.show_replaced
    Crafty_sv.show_replaced = crafty.show_replaced
    printfh("Show replacements: <font color=\"#ffffcc\">%s</font>", tostring(crafty.show_replaced))
    did_something = true
  end

  if args.s then
    crafty.toggle_ui()
    -- we temporarily disable the auto-show, without altering
    -- the setting in saved variables.
    crafty.disable_quest_hooks()
    crafty.autoshow = false
    did_something = true
  else
    -- force an update.
    -- only do this if we aren't forcing the display visible anyway.
    if args.f or args.m or args.r or args.v or args.x or args.y then
      if crafty.ui and crafty.ui:GetVisible() then
        crafty.update_hook()
      end
    end
  end
  if not did_something then
    local lines = crafty.display_needed(crafty.compute())
    printfhs(table.concat(lines, "\n  "))
  end
  if crafty.autoshow then
    crafty.enable_quest_hooks()
  end
end

Library.LibGetOpt.makeslash("adf#m#rsvVx#y#", "Crafty", "crafty", crafty.slashcommand)
Command.Event.Attach(Event.Addon.SavedVariables.Load.End, crafty.variables, "variable check")
